#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "common.h"

#define _PATH_DEV_DM		"/dev/mapper"
#define _PATH_DEV_MPATH		"/dev/mpath"

/*
 * return partition name - uses static storage unless buf is supplied
 */
static char *
partnamebf(char *dev, int pno, int lth, int bufsiz, char *bufp) {
	static char buffer[80];
	char *p;
	int w, wp;

	if (!bufp) {
		bufp = buffer;
		bufsiz = sizeof(buffer);
	}

	w = strlen(dev);
	p = "";

	if (isdigit(dev[w-1]))
		p = "p";

	/* devfs kludge - note: fdisk partition names are not supposed
	   to equal kernel names, so there is no reason to do this */
	if (strcmp (dev + w - 4, "disc") == 0) {
		w -= 4;
		p = "part";
	}

	/* /dev/mapper and /dev/mpath usually use "p<N>"
	 * (note that this is not a strict rule, the suffix is possible to define
	 *  on kpartx command line. This code follows usual behavior.)
	 */
	if (strncmp(dev, _PATH_DEV_DM, strlen(_PATH_DEV_DM)) == 0 ||
	    strncmp(dev, _PATH_DEV_MPATH, strlen(_PATH_DEV_MPATH)) == 0)
		p = "p";

	wp = strlen(p);

	if (lth) {
		snprintf(bufp, bufsiz, "%*.*s%s%-2u",
			 lth-wp-2, w, dev, p, pno);
	} else {
		snprintf(bufp, bufsiz, "%.*s%s%-2u", w, dev, p, pno);
	}
	return bufp;
}

char *
partname(char *dev, int pno, int lth) {
	return partnamebf(dev, pno, lth, 0, NULL);
}

int
is_probably_full_disk(char *name) {
#ifdef HDIO_GETGEO
	struct hd_geometry geometry;
	int fd, i = 0;

	fd = open(name, O_RDONLY);
	if (fd >= 0) {
		i = ioctl(fd, HDIO_GETGEO, &geometry);
		close(fd);
	}
	if (i==0)
		return (fd >= 0 && geometry.start == 0);
	/*
	 * "silly heuristic" is still so sexy for us, because
	 * for example Xen doesn't implement HDIO_GETGEO for virtual
	 * block devices (/dev/xvda).
	 * -- kzak@redhat.com (23-Feb-2006)
	 */
#endif
	/* silly heuristic */
	while (*name)
		name++;
	return !isdigit(name[-1]);
}

